
/*
 * renderer.js v1.5.2
 * (c) 2017 @Johnny Wu
 * Released under the MIT License.
 */

'use strict';

function _interopDefault (ex) { return (ex && (typeof ex === 'object') && 'default' in ex) ? ex['default'] : ex; }

var gfx = _interopDefault(require('gfx.js'));
var vmath = require('vmath');
var memop = require('memop');

var enums = {
  // projection
  PROJ_PERSPECTIVE: 0,
  PROJ_ORTHO: 1,

  // lights
  LIGHT_DIRECTIONAL: 0,
  LIGHT_POINT: 1,
  LIGHT_SPOT: 2,

  // parameter type
  PARAM_INT:             0,
  PARAM_INT2:            1,
  PARAM_INT3:            2,
  PARAM_INT4:            3,
  PARAM_FLOAT:           4,
  PARAM_FLOAT2:          5,
  PARAM_FLOAT3:          6,
  PARAM_FLOAT4:          7,
  PARAM_COLOR3:          8,
  PARAM_COLOR4:          9,
  PARAM_MAT2:           10,
  PARAM_MAT3:           11,
  PARAM_MAT4:           12,
  PARAM_TEXTURE_2D:     13,
  PARAM_TEXTURE_CUBE:   14,

  // clear flags
  CLEAR_COLOR: 1,
  CLEAR_DEPTH: 2,
  CLEAR_STENCIL: 4,
};

var InputAssembler = function InputAssembler(vb, ib, pt) {
  if ( pt === void 0 ) pt = gfx.PT_TRIANGLES;

  this._vertexBuffer = vb;
  this._indexBuffer = ib;
  this._primitiveType = pt;
  this._start = 0;
  this._count = -1;

  // TODO: instancing data
  // this._stream = 0;
};

InputAssembler.prototype.getPrimitiveCount = function getPrimitiveCount () {
  if (this._count !== -1) {
    return this._count;
  }

  if (this._indexBuffer) {
    return this._indexBuffer.count;
  }

  return this._vertexBuffer.count;
};

var Pass = function Pass(name) {
  this._programName = name;

  // cullmode
  this._cullMode = gfx.CULL_BACK;

  // blending
  this._blend = false;
  this._blendEq = gfx.BLEND_FUNC_ADD;
  this._blendAlphaEq = gfx.BLEND_FUNC_ADD;
  this._blendSrc = gfx.BLEND_ONE;
  this._blendDst = gfx.BLEND_ZERO;
  this._blendSrcAlpha = gfx.BLEND_ONE;
  this._blendDstAlpha = gfx.BLEND_ZERO;
  this._blendColor = 0xffffffff;

  // depth
  this._depthTest = false;
  this._depthWrite = false;
  this._depthFunc = gfx.DS_FUNC_LESS,

  // stencil
  this._stencilTest = false;
  // front
  this._stencilFuncFront = gfx.DS_FUNC_ALWAYS;
  this._stencilRefFront = 0;
  this._stencilMaskFront = 0xff;
  this._stencilFailOpFront = gfx.STENCIL_OP_KEEP;
  this._stencilZFailOpFront = gfx.STENCIL_OP_KEEP;
  this._stencilZPassOpFront = gfx.STENCIL_OP_KEEP;
  this._stencilWriteMaskFront = 0xff;
  // back
  this._stencilFuncBack = gfx.DS_FUNC_ALWAYS;
  this._stencilRefBack = 0;
  this._stencilMaskBack = 0xff;
  this._stencilFailOpBack = gfx.STENCIL_OP_KEEP;
  this._stencilZFailOpBack = gfx.STENCIL_OP_KEEP;
  this._stencilZPassOpBack = gfx.STENCIL_OP_KEEP;
  this._stencilWriteMaskBack = 0xff;
};

Pass.prototype.setCullMode = function setCullMode (cullMode) {
  this._cullMode = cullMode;
};

Pass.prototype.setBlend = function setBlend (
  blendEq,
  blendSrc,
  blendDst,
  blendAlphaEq,
  blendSrcAlpha,
  blendDstAlpha,
  blendColor
) {
    if ( blendEq === void 0 ) blendEq = gfx.BLEND_FUNC_ADD;
    if ( blendSrc === void 0 ) blendSrc = gfx.BLEND_ONE;
    if ( blendDst === void 0 ) blendDst = gfx.BLEND_ZERO;
    if ( blendAlphaEq === void 0 ) blendAlphaEq = gfx.BLEND_FUNC_ADD;
    if ( blendSrcAlpha === void 0 ) blendSrcAlpha = gfx.BLEND_ONE;
    if ( blendDstAlpha === void 0 ) blendDstAlpha = gfx.BLEND_ZERO;
    if ( blendColor === void 0 ) blendColor = 0xffffffff;

  this._blend = true;
  this._blendEq = blendEq;
  this._blendSrc = blendSrc;
  this._blendDst = blendDst;
  this._blendAlphaEq = blendAlphaEq;
  this._blendSrcAlpha = blendSrcAlpha;
  this._blendDstAlpha = blendDstAlpha;
  this._blendColor = blendColor;
};

Pass.prototype.setDepth = function setDepth (
  depthTest,
  depthWrite,
  depthFunc
) {
    if ( depthTest === void 0 ) depthTest = false;
    if ( depthWrite === void 0 ) depthWrite = false;
    if ( depthFunc === void 0 ) depthFunc = gfx.DS_FUNC_LESS;

  this._depthTest = depthTest;
  this._depthWrite = depthWrite;
  this._depthFunc = depthFunc;
};

Pass.prototype.setStencilFront = function setStencilFront (
  stencilFunc,
  stencilRef,
  stencilMask,
  stencilFailOp,
  stencilZFailOp,
  stencilZPassOp,
  stencilWriteMask
) {
    if ( stencilFunc === void 0 ) stencilFunc = gfx.DS_FUNC_ALWAYS;
    if ( stencilRef === void 0 ) stencilRef = 0;
    if ( stencilMask === void 0 ) stencilMask = 0xff;
    if ( stencilFailOp === void 0 ) stencilFailOp = gfx.STENCIL_OP_KEEP;
    if ( stencilZFailOp === void 0 ) stencilZFailOp = gfx.STENCIL_OP_KEEP;
    if ( stencilZPassOp === void 0 ) stencilZPassOp = gfx.STENCIL_OP_KEEP;
    if ( stencilWriteMask === void 0 ) stencilWriteMask = 0xff;

  this._stencilTest = true;
  this._stencilFuncFront = stencilFunc;
  this._stencilRefFront = stencilRef;
  this._stencilMaskFront = stencilMask;
  this._stencilFailOpFront = stencilFailOp;
  this._stencilZFailOpFront = stencilZFailOp;
  this._stencilZPassOpFront = stencilZPassOp;
  this._stencilWriteMaskFront = stencilWriteMask;
};

Pass.prototype.setStencilBack = function setStencilBack (
  stencilFunc,
  stencilRef,
  stencilMask,
  stencilFailOp,
  stencilZFailOp,
  stencilZPassOp,
  stencilWriteMask
) {
    if ( stencilFunc === void 0 ) stencilFunc = gfx.DS_FUNC_ALWAYS;
    if ( stencilRef === void 0 ) stencilRef = 0;
    if ( stencilMask === void 0 ) stencilMask = 0xff;
    if ( stencilFailOp === void 0 ) stencilFailOp = gfx.STENCIL_OP_KEEP;
    if ( stencilZFailOp === void 0 ) stencilZFailOp = gfx.STENCIL_OP_KEEP;
    if ( stencilZPassOp === void 0 ) stencilZPassOp = gfx.STENCIL_OP_KEEP;
    if ( stencilWriteMask === void 0 ) stencilWriteMask = 0xff;

  this._stencilTest = true;
  this._stencilFuncBack = stencilFunc;
  this._stencilRefBack = stencilRef;
  this._stencilMaskBack = stencilMask;
  this._stencilFailOpBack = stencilFailOp;
  this._stencilZFailOpBack = stencilZFailOp;
  this._stencilZPassOpBack = stencilZPassOp;
  this._stencilWriteMaskBack = stencilWriteMask;
};

var _stageOffset = 0;
var _name2stageID = {};

var config = {
  addStage: function (name) {
    // already added
    if (_name2stageID[name] !== undefined) {
      return;
    }

    var stageID = 1 << _stageOffset;
    _name2stageID[name] = stageID;

    _stageOffset += 1;
  },

  stageID: function (name) {
    var id = _name2stageID[name];
    if (id === undefined) {
      return -1;
    }
    return id;
  },

  stageIDs: function (nameList) {
    var key = 0;
    for (var i = 0; i < nameList.length; ++i) {
      var id = _name2stageID[nameList[i]];
      if (id !== undefined) {
        key |= id;
      }
    }
    return key;
  }
};

var _genID = 0;

var Technique = function Technique(stages, parameters, passes, layer) {
  if ( layer === void 0 ) layer = 0;

  this._id = _genID++;
  this._stageIDs = config.stageIDs(stages);
  this._parameters = parameters; // {name, type, size, val}
  this._passes = passes;
  this._layer = layer;
  // TODO: this._version = 'webgl' or 'webgl2' // ????
};

var prototypeAccessors = { passes: {},stageIDs: {} };

Technique.prototype.setStages = function setStages (stages) {
  this._stageIDs = config.stageIDs(stages);
};

prototypeAccessors.passes.get = function () {
  return this._passes;
};

prototypeAccessors.stageIDs.get = function () {
  return this._stageIDs;
};

Object.defineProperties( Technique.prototype, prototypeAccessors );

var Effect = function Effect(techniques, values, opts) {
  if ( values === void 0 ) values = {};
  if ( opts === void 0 ) opts = [];

  this._techniques = techniques;
  this._values = values;
  this._options = opts;

  // TODO: check if params is valid for current technique???
};

Effect.prototype.getTechnique = function getTechnique (stage) {
    var this$1 = this;

  var stageID = config.stageID(stage);
  for (var i = 0; i < this._techniques.length; ++i) {
    var tech = this$1._techniques[i];
    if (tech.stageIDs & stageID) {
      return tech;
    }
  }

  return null;
};

Effect.prototype.getValue = function getValue (name) {
  return this._values[name];
};

Effect.prototype.setValue = function setValue (name, value) {
  // TODO: check if params is valid for current technique???
  this._values[name] = value;
};

Effect.prototype.getOption = function getOption (name) {
    var this$1 = this;

  for (var i = 0; i < this._options.length; ++i) {
    var opt = this$1._options[i];
    if ( opt.name === name ) {
      return opt.value;
    }
  }

  console.warn(("Failed to get option " + name + ", option not found."));
  return null;
};

Effect.prototype.setOption = function setOption (name, value) {
    var this$1 = this;

  for (var i = 0; i < this._options.length; ++i) {
    var opt = this$1._options[i];
    if ( opt.name === name ) {
      opt.value = value;
      return;
    }
  }

  console.warn(("Failed to set option " + name + ", option not found."));
};

Effect.prototype.extractOptions = function extractOptions (out) {
    var this$1 = this;
    if ( out === void 0 ) out = {};

  for (var i = 0; i < this._options.length; ++i) {
    var opt = this$1._options[i];
    out[opt.name] = opt.value;
  }

  return out;
};

/**
 * @param {object} json
 */


/**
 * @param {gfx.Device} device
 * @param {Object} data
 */
function createIA(device, data) {
  if (!data.positions) {
    console.error('The data must have positions field');
    return null;
  }

  var verts = [];
  var vcount = data.positions.length / 3;

  for (var i = 0; i < vcount; ++i) {
    verts.push(data.positions[3 * i], data.positions[3 * i + 1], data.positions[3 * i + 2]);

    if (data.normals) {
      verts.push(data.normals[3 * i], data.normals[3 * i + 1], data.normals[3 * i + 2]);
    }

    if (data.uvs) {
      verts.push(data.uvs[2 * i], data.uvs[2 * i + 1]);
    }
  }

  var vfmt = [];
  vfmt.push({ name: gfx.ATTR_POSITION, type: gfx.ATTR_TYPE_FLOAT32, num: 3 });
  if (data.normals) {
    vfmt.push({ name: gfx.ATTR_NORMAL, type: gfx.ATTR_TYPE_FLOAT32, num: 3 });
  }
  if (data.uvs) {
    vfmt.push({ name: gfx.ATTR_UV0, type: gfx.ATTR_TYPE_FLOAT32, num: 2 });
  }

  var vb = new gfx.VertexBuffer(
    device,
    new gfx.VertexFormat(vfmt),
    gfx.USAGE_STATIC,
    new Float32Array(verts),
    vcount
  );

  var ib = null;
  if (data.indices) {
    ib = new gfx.IndexBuffer(
      device,
      gfx.INDEX_FMT_UINT16,
      gfx.USAGE_STATIC,
      new Uint16Array(data.indices),
      data.indices.length
    );
  }

  return new InputAssembler(vb, ib);
}

var _m4_tmp = vmath.mat4.create();
var _genID$1 = 0;

var View = function View() {
  this._id = _genID$1++;

  // viewport
  this._rect = {
    x: 0, y: 0, w: 1, h: 1
  };

  // TODO:
  // this._scissor = {
  // x: 0, y: 0, w: 1, h: 1
  // };

  // clear options
  this._color = vmath.color4.new(0.3, 0.3, 0.3, 1);
  this._depth = 1;
  this._stencil = 1;
  this._clearFlags = enums.CLEAR_COLOR | enums.CLEAR_DEPTH;

  // matrix
  this._matView = vmath.mat4.create();
  this._matProj = vmath.mat4.create();
  this._matViewProj = vmath.mat4.create();
  this._matInvViewProj = vmath.mat4.create();

  // stages
  this._stages = [];
  this._cullingByID = false;
};

View.prototype.getForward = function getForward (out) {
  return vmath.vec3.set(
    out,
    -this._matView.m02,
    -this._matView.m06,
    -this._matView.m10
  );
};

View.prototype.getPosition = function getPosition (out) {
  vmath.mat4.invert(_m4_tmp, this._matView);
  return vmath.mat4.getTranslation(out, _m4_tmp);
};

var _m4_tmp$1 = vmath.mat4.create();
var _m3_tmp = vmath.mat3.create();
var _forward = vmath.vec3.new(0, 0, -1);
var _transformedLightDirection = vmath.vec3.create();

var Light = function Light() {
  this._poolID = -1;
  this._node = null;

  this._type = enums.LIGHT_DIRECTIONAL;

  this._color = vmath.color3.new(1, 1, 1);
  this._intensity = 1;

  // used for spot and point light
  this._range = 1;
  // used for spot light, default to 60 degrees
  this._spotAngle = vmath.toRadian(60);
  this._spotExp = 1;
  // cached for uniform
  this._directionUniform = new Float32Array(3);
  this._positionUniform = new Float32Array(3);
  this._colorUniform = new Float32Array([this._color.r * this._intensity, this._color.g * this._intensity, this._color.b * this._intensity]);
  this._spotUniform = new Float32Array([Math.cos(this._spotAngle * 0.5), this._spotExp]);
};

var prototypeAccessors$1 = { color: {},intensity: {},type: {},spotAngle: {},spotExp: {},range: {} };

Light.prototype.setNode = function setNode (node) {
  this._node = node;
};

prototypeAccessors$1.color.set = function (val) {
  vmath.color3.copy(this._color, val);
  this._colorUniform[0] = val.r * this._intensity;
  this._colorUniform[1] = val.g * this._intensity;
  this._colorUniform[2] = val.b * this._intensity;
};
prototypeAccessors$1.color.get = function () {
  return this._color;
};

prototypeAccessors$1.intensity.set = function (val) {
  this._intensity = val;
  this._colorUniform[0] = val * this._color.r;
  this._colorUniform[1] = val * this._color.g;
  this._colorUniform[2] = val * this._color.b;
};
prototypeAccessors$1.intensity.get = function () {
  return this._intensity;
};

prototypeAccessors$1.type.set = function (tpe) {
  this._type = tpe;
};
prototypeAccessors$1.type.get = function () {
  return this._type;
};

prototypeAccessors$1.spotAngle.set = function (val) {
  this._spotAngle = val;
  this._spotUniform[0] = Math.cos(this._spotAngle * 0.5);
};
prototypeAccessors$1.spotAngle.get = function () {
  return this._spotAngle;
};

prototypeAccessors$1.spotExp.set = function (val) {
  this._spotExp = val;
  this._spotUniform[1] = val;
};
prototypeAccessors$1.spotExp.get = function () {
  return this._spotExp;
};

prototypeAccessors$1.range.set = function (tpe) {
  this._range = tpe;
};
prototypeAccessors$1.range.get = function () {
  return this._range;
};

Light.prototype._updateLightPositionAndDirection = function _updateLightPositionAndDirection () {
  this._node.getWorldMatrix(_m4_tmp$1);
  vmath.mat3.fromMat4(_m3_tmp, _m4_tmp$1);
  vmath.vec3.transformMat3(_transformedLightDirection, _forward, _m3_tmp);
  vmath.vec3.array(this._directionUniform, _transformedLightDirection);
  var pos = this._positionUniform;
  pos[0] = _m4_tmp$1.m12;
  pos[1] = _m4_tmp$1.m13;
  pos[2] = _m4_tmp$1.m14;
};

Light.prototype.update = function update () {
  this._updateLightPositionAndDirection();
};

Object.defineProperties( Light.prototype, prototypeAccessors$1 );

var _matView = vmath.mat4.create();
var _matProj = vmath.mat4.create();
var _matViewProj = vmath.mat4.create();
var _matInvViewProj = vmath.mat4.create();
var _tmp_v3 = vmath.vec3.create();

var Camera = function Camera() {
  this._poolID = -1;
  this._node = null;

  //
  this._projection = enums.PROJ_PERSPECTIVE;

  // clear options
  this._color = vmath.color4.new(0.2, 0.3, 0.47, 1);
  this._depth = 1;
  this._stencil = 1;
  this._clearFlags = enums.CLEAR_COLOR | enums.CLEAR_DEPTH;

  // projection properties
  this._near = 0.01;
  this._far = 1000.0;
  this._fov = Math.PI/4.0; // vertical fov
  // this._aspect = 16.0/9.0; // DISABLE: use _rect.w/_rect.h
  this._rect = {
    x: 0, y: 0, w: 1, h: 1
  };

  // ortho properties
  this._orthoHeight = 10;
  this._stages = [];
};

Camera.prototype.setColor = function setColor (r, g, b, a) {
  vmath.color4.set(this._color, r, g, b, a);
};

Camera.prototype.setDepth = function setDepth (depth) {
  this._depth = depth;
};

Camera.prototype.setStencil = function setStencil (stencil) {
  this._stencil = stencil;
};

Camera.prototype.setClearFlags = function setClearFlags (flags) {
  this._clearFlags = flags;
};

Camera.prototype.setNode = function setNode (node) {
  this._node = node;
};

/**
 * @param {Number} x - [0,1]
 * @param {Number} y - [0,1]
 * @param {Number} w - [0,1]
 * @param {Number} h - [0,1]
 */
Camera.prototype.setRect = function setRect (x, y, w, h) {
  this._rect.x = x;
  this._rect.y = y;
  this._rect.w = w;
  this._rect.h = h;
};

Camera.prototype.setStages = function setStages (stages) {
  this._stages = stages;
};

Camera.prototype.extractView = function extractView (out, width, height) {
  // rect
  out._rect.x = this._rect.x * width;
  out._rect.y = this._rect.y * height;
  out._rect.w = this._rect.w * width;
  out._rect.h = this._rect.h * height;

  // clear opts
  out._color = this._color;
  out._depth = this._depth;
  out._stencil = this._stencil;
  out._clearFlags = this._clearFlags;

  // view matrix
  this._node.getWorldRT(out._matView);
  vmath.mat4.invert(out._matView, out._matView);

  // projection matrix
  // TODO: if this._projDirty
  var aspect = width / height;
  if (this._projection === enums.PROJ_PERSPECTIVE) {
    vmath.mat4.perspective(out._matProj,
      this._fov,
      aspect,
      this._near,
      this._far
    );
  } else {
    var x = this._orthoHeight * aspect;
    var y = this._orthoHeight;
    vmath.mat4.ortho(out._matProj,
      -x, x, -y, y, this._near, this._far
    );
  }

  // view-projection
  vmath.mat4.mul(out._matViewProj, out._matProj, out._matView);
  vmath.mat4.invert(out._matInvViewProj, out._matViewProj);

  // stages
  out._stages = this._stages;
};

Camera.prototype.screenToWorld = function screenToWorld (out, screenPos, width, height) {
  var aspect = width / height;
  var cx = this._rect.x * width;
  var cy = this._rect.y * height;
  var cw = this._rect.w * width;
  var ch = this._rect.h * height;

  // view matrix
  this._node.getWorldRT(_matView);
  vmath.mat4.invert(_matView, _matView);

  // projection matrix
  if (this._projection === enums.PROJ_PERSPECTIVE) {
    vmath.mat4.perspective(_matProj,
      this._fov,
      aspect,
      this._near,
      this._far
    );
  } else {
    var x = this._orthoHeight * aspect;
    var y = this._orthoHeight;
    vmath.mat4.ortho(_matProj,
      -x, x, -y, y, this._near, this._far
    );
  }

  // view-projection
  vmath.mat4.mul(_matViewProj, _matProj, _matView);

  // inv view-projection
  vmath.mat4.invert(_matInvViewProj, _matViewProj);

  //
  if (this._projection === enums.PROJ_PERSPECTIVE) {
    // calculate screen pos in far clip plane
    vmath.vec3.set(out,
      (screenPos.x - cx) * 2.0 / cw - 1.0,
      (screenPos.y - cy) * 2.0 / ch - 1.0, // DISABLE: (ch - (screenPos.y - cy)) * 2.0 / ch - 1.0,
      1.0
    );

    // transform to world
    vmath.vec3.transformMat4(out, out, _matInvViewProj);

    //
    this._node.getWorldPos(_tmp_v3);
    vmath.vec3.lerp(out, _tmp_v3, out, screenPos.z / this._far);
  } else {
    var range = this._farClip - this._nearClip;
    vmath.vec3.set(out,
      (screenPos.x - cx) * 2.0 / cw - 1.0,
      (screenPos.y - cy) * 2.0 / ch - 1.0, // DISABLE: (ch - (screenPos.y - cy)) * 2.0 / ch - 1.0,
      (this._far - screenPos.z) / range * 2.0 - 1.0
    );

    // transform to world
    vmath.vec3.transformMat4(out, out, _matInvViewProj);
  }

  return out;
};

Camera.prototype.worldToScreen = function worldToScreen (out, worldPos, width, height) {
  var aspect = width / height;
  var cx = this._rect.x * width;
  var cy = this._rect.y * height;
  var cw = this._rect.w * width;
  var ch = this._rect.h * height;

  // view matrix
  this._node.getWorldRT(_matView);
  vmath.mat4.invert(_matView, _matView);

  // projection matrix
  if (this._projection === enums.PROJ_PERSPECTIVE) {
    vmath.mat4.perspective(_matProj,
      this._fov,
      aspect,
      this._near,
      this._far
    );
  } else {
    var x = this._orthoHeight * aspect;
    var y = this._orthoHeight;
    vmath.mat4.ortho(_matProj,
      -x, x, -y, y, this._near, this._far
    );
  }

  // view-projection
  vmath.mat4.mul(_matViewProj, _matProj, _matView);

  // calculate w
  var w =
    worldPos.x * _matViewProj.m03 +
    worldPos.y * _matViewProj.m07 +
    worldPos.z * _matViewProj.m11 +
    _matViewProj.m15;

  vmath.vec3.transformMat4(out, worldPos, _matViewProj);
  out.x = cx + (out.x / w + 1) * 0.5 * cw;
  out.y = cy + (out.y / w + 1) * 0.5 * ch;

  return out;
};

var Model = function Model() {
  this._poolID = -1;
  this._node = null;
  this._inputAssemblers = [];
  this._effects = [];
  this._options = [];
  this._dynamicIA = false;
  this._viewID = -1;

  // TODO: we calculate aabb based on vertices
  // this._aabb
};

var prototypeAccessors$2 = { inputAssemblerCount: {},dynamicIA: {},drawItemCount: {} };

prototypeAccessors$2.inputAssemblerCount.get = function () {
  return this._inputAssemblers.length;
};

prototypeAccessors$2.dynamicIA.get = function () {
  return this._dynamicIA;
};

prototypeAccessors$2.drawItemCount.get = function () {
  return this._dynamicIA ? 1 : this._inputAssemblers.length;
};

Model.prototype.setNode = function setNode (node) {
  this._node = node;
};

Model.prototype.setDynamicIA = function setDynamicIA (enabled) {
  this._dynamicIA = enabled;
};

Model.prototype.addInputAssembler = function addInputAssembler (ia) {
  if (this._inputAssemblers.indexOf(ia) !== -1) {
    return;
  }
  this._inputAssemblers.push(ia);
};

Model.prototype.clearInputAssemblers = function clearInputAssemblers () {
  this._inputAssemblers.length = 0;
};

Model.prototype.addEffect = function addEffect (effect) {
  if (this._effects.indexOf(effect) !== -1) {
    return;
  }
  this._effects.push(effect);

  //
  var opts = Object.create(null);
  effect.extractOptions(opts);
  this._options.push(opts);
};

Model.prototype.clearEffects = function clearEffects () {
  this._effects.length = 0;
  this._options.length = 0;
};

Model.prototype.extractDrawItem = function extractDrawItem (out, index) {
  if (this._dynamicIA) {
    out.model = this;
    out.node = this._node;
    out.ia = null;
    out.effect = this._effects[0];
    out.options = out.effect.extractOptions(this._options[0]);

    return;
  }

  if (index >= this._inputAssemblers.length ) {
    out.model = null;
    out.node = null;
    out.ia = null;
    out.effect = null;
    out.options = null;

    return;
  }

  out.model = this;
  out.node = this._node;
  out.ia = this._inputAssemblers[index];

  var effect, options;
  if (index < this._effects.length) {
    effect = this._effects[index];
    options = this._options[index];
  } else {
    effect = this._effects[this._effects.length-1];
    options = this._options[this._effects.length-1];
  }
  out.effect = effect;
  out.options = effect.extractOptions(options);
};

Object.defineProperties( Model.prototype, prototypeAccessors$2 );

var Scene = function Scene() {
  this._lights = new memop.FixedArray(16);
  this._models = new memop.FixedArray(16);
  this._cameras = new memop.FixedArray(16);
  this._debugCamera = null;

  // NOTE: we don't use pool for views (because it's less changed and it doesn't have poolID)
  this._views = [];
};

Scene.prototype._add = function _add (pool, item) {
  if (item._poolID !== -1) {
    return;
  }

  pool.push(item);
  item._poolID = pool.length - 1;
};

Scene.prototype._remove = function _remove (pool, item) {
  if (item._poolID === -1) {
    return;
  }

  pool.data[pool.length-1]._poolID = item._poolID;
  pool.fastRemove(item._poolID);
  item._poolID = -1;
};

Scene.prototype.reset = function reset () {
    var this$1 = this;

  for (var i = 0; i < this._models.length; ++i) {
    var model = this$1._models.data[i];
    model._viewID = -1;
  }
};

Scene.prototype.setDebugCamera = function setDebugCamera (cam) {
  this._debugCamera = cam;
};

// camera

Scene.prototype.getCameraCount = function getCameraCount () {
  return this._cameras.length;
};

Scene.prototype.getCamera = function getCamera (idx) {
  return this._cameras.data[idx];
};

Scene.prototype.addCamera = function addCamera (camera) {
  this._add(this._cameras, camera);
};

Scene.prototype.removeCamera = function removeCamera (camera) {
  this._remove(this._cameras, camera);
};

// model

Scene.prototype.getModelCount = function getModelCount () {
  return this._models.length;
};

Scene.prototype.getModel = function getModel (idx) {
  return this._models.data[idx];
};

Scene.prototype.addModel = function addModel (model) {
  this._add(this._models, model);
};

Scene.prototype.removeModel = function removeModel (model) {
  this._remove(this._models, model);
};

// light

Scene.prototype.getLightCount = function getLightCount () {
  return this._lights.length;
};

Scene.prototype.getLight = function getLight (idx) {
  return this._lights.data[idx];
};

Scene.prototype.addLight = function addLight (light) {
  this._add(this._lights, light);
};

Scene.prototype.removeLight = function removeLight (light) {
  this._remove(this._lights, light);
};

// view

Scene.prototype.addView = function addView (view) {
  if (this._views.indexOf(view) === -1) {
    this._views.push(view);
  }
};

Scene.prototype.removeView = function removeView (view) {
  var idx = this._views.indexOf(view);
  if (idx !== -1) {
    this._views.splice(idx, 1);
  }
};

/*!
 * mustache.js - Logic-less {{mustache}} templates with JavaScript
 * http://github.com/janl/mustache.js
 */

var mustache = {};

var objectToString = Object.prototype.toString;
var isArray = Array.isArray || function isArrayPolyfill(object) {
  return objectToString.call(object) === '[object Array]';
};

function isFunction(object) {
  return typeof object === 'function';
}

/**
 * More correct typeof string handling array
 * which normally returns typeof 'object'
 */
function typeStr(obj) {
  return isArray(obj) ? 'array' : typeof obj;
}

function escapeRegExp(string) {
  return string.replace(/[\-\[\]{}()*+?.,\\\^$|#\s]/g, '\\$&');
}

/**
 * Null safe way of checking whether or not an object,
 * including its prototype, has a given property
 */
function hasProperty(obj, propName) {
  return obj != null && typeof obj === 'object' && (propName in obj);
}

// Workaround for https://issues.apache.org/jira/browse/COUCHDB-577
// See https://github.com/janl/mustache.js/issues/189
var regExpTest = RegExp.prototype.test;
function testRegExp(re, string) {
  return regExpTest.call(re, string);
}

var nonSpaceRe = /\S/;
function isWhitespace(string) {
  return !testRegExp(nonSpaceRe, string);
}

var entityMap = {
  '&': '&amp;',
  '<': '&lt;',
  '>': '&gt;',
  '"': '&quot;',
  "'": '&#39;',
  '/': '&#x2F;',
  '`': '&#x60;',
  '=': '&#x3D;'
};

function escapeHtml(string) {
  return String(string).replace(/[&<>"'`=\/]/g, function fromEntityMap(s) {
    return entityMap[s];
  });
}

var whiteRe = /\s*/;
var spaceRe = /\s+/;
var equalsRe = /\s*=/;
var curlyRe = /\s*\}/;
var tagRe = /#|\^|\/|>|\{|&|=|!/;

/**
 * Breaks up the given `template` string into a tree of tokens. If the `tags`
 * argument is given here it must be an array with two string values: the
 * opening and closing tags used in the template (e.g. [ "<%", "%>" ]). Of
 * course, the default is to use mustaches (i.e. mustache.tags).
 *
 * A token is an array with at least 4 elements. The first element is the
 * mustache symbol that was used inside the tag, e.g. "#" or "&". If the tag
 * did not contain a symbol (i.e. {{myValue}}) this element is "name". For
 * all text that appears outside a symbol this element is "text".
 *
 * The second element of a token is its "value". For mustache tags this is
 * whatever else was inside the tag besides the opening symbol. For text tokens
 * this is the text itself.
 *
 * The third and fourth elements of the token are the start and end indices,
 * respectively, of the token in the original template.
 *
 * Tokens that are the root node of a subtree contain two more elements: 1) an
 * array of tokens in the subtree and 2) the index in the original template at
 * which the closing tag for that section begins.
 */
function parseTemplate(template, tags) {
  if (!template)
    { return []; }

  var sections = [];     // Stack to hold section tokens
  var tokens = [];       // Buffer to hold the tokens
  var spaces = [];       // Indices of whitespace tokens on the current line
  var hasTag = false;    // Is there a {{tag}} on the current line?
  var nonSpace = false;  // Is there a non-space char on the current line?

  // Strips all whitespace tokens array for the current line
  // if there was a {{#tag}} on it and otherwise only space.
  function stripSpace() {
    if (hasTag && !nonSpace) {
      while (spaces.length)
        { delete tokens[spaces.pop()]; }
    } else {
      spaces = [];
    }

    hasTag = false;
    nonSpace = false;
  }

  var openingTagRe, closingTagRe, closingCurlyRe;
  function compileTags(tagsToCompile) {
    if (typeof tagsToCompile === 'string')
      { tagsToCompile = tagsToCompile.split(spaceRe, 2); }

    if (!isArray(tagsToCompile) || tagsToCompile.length !== 2)
      { throw new Error('Invalid tags: ' + tagsToCompile); }

    openingTagRe = new RegExp(escapeRegExp(tagsToCompile[0]) + '\\s*');
    closingTagRe = new RegExp('\\s*' + escapeRegExp(tagsToCompile[1]));
    closingCurlyRe = new RegExp('\\s*' + escapeRegExp('}' + tagsToCompile[1]));
  }

  compileTags(tags || mustache.tags);

  var scanner = new Scanner(template);

  var start, type, value, chr, token, openSection;
  while (!scanner.eos()) {
    start = scanner.pos;

    // Match any text between tags.
    value = scanner.scanUntil(openingTagRe);

    if (value) {
      for (var i = 0, valueLength = value.length; i < valueLength; ++i) {
        chr = value.charAt(i);

        if (isWhitespace(chr)) {
          spaces.push(tokens.length);
        } else {
          nonSpace = true;
        }

        tokens.push(['text', chr, start, start + 1]);
        start += 1;

        // Check for whitespace on the current line.
        if (chr === '\n')
          { stripSpace(); }
      }
    }

    // Match the opening tag.
    if (!scanner.scan(openingTagRe))
      { break; }

    hasTag = true;

    // Get the tag type.
    type = scanner.scan(tagRe) || 'name';
    scanner.scan(whiteRe);

    // Get the tag value.
    if (type === '=') {
      value = scanner.scanUntil(equalsRe);
      scanner.scan(equalsRe);
      scanner.scanUntil(closingTagRe);
    } else if (type === '{') {
      value = scanner.scanUntil(closingCurlyRe);
      scanner.scan(curlyRe);
      scanner.scanUntil(closingTagRe);
      type = '&';
    } else {
      value = scanner.scanUntil(closingTagRe);
    }

    // Match the closing tag.
    if (!scanner.scan(closingTagRe))
      { throw new Error('Unclosed tag at ' + scanner.pos); }

    token = [type, value, start, scanner.pos];
    tokens.push(token);

    if (type === '#' || type === '^') {
      sections.push(token);
    } else if (type === '/') {
      // Check section nesting.
      openSection = sections.pop();

      if (!openSection)
        { throw new Error('Unopened section "' + value + '" at ' + start); }

      if (openSection[1] !== value)
        { throw new Error('Unclosed section "' + openSection[1] + '" at ' + start); }
    } else if (type === 'name' || type === '{' || type === '&') {
      nonSpace = true;
    } else if (type === '=') {
      // Set the tags for the next time around.
      compileTags(value);
    }
  }

  // Make sure there are no open sections when we're done.
  openSection = sections.pop();

  if (openSection)
    { throw new Error('Unclosed section "' + openSection[1] + '" at ' + scanner.pos); }

  return nestTokens(squashTokens(tokens));
}

/**
 * Combines the values of consecutive text tokens in the given `tokens` array
 * to a single token.
 */
function squashTokens(tokens) {
  var squashedTokens = [];

  var token, lastToken;
  for (var i = 0, numTokens = tokens.length; i < numTokens; ++i) {
    token = tokens[i];

    if (token) {
      if (token[0] === 'text' && lastToken && lastToken[0] === 'text') {
        lastToken[1] += token[1];
        lastToken[3] = token[3];
      } else {
        squashedTokens.push(token);
        lastToken = token;
      }
    }
  }

  return squashedTokens;
}

/**
 * Forms the given array of `tokens` into a nested tree structure where
 * tokens that represent a section have two additional items: 1) an array of
 * all tokens that appear in that section and 2) the index in the original
 * template that represents the end of that section.
 */
function nestTokens(tokens) {
  var nestedTokens = [];
  var collector = nestedTokens;
  var sections = [];

  var token, section;
  for (var i = 0, numTokens = tokens.length; i < numTokens; ++i) {
    token = tokens[i];

    switch (token[0]) {
      case '#':
      case '^':
        collector.push(token);
        sections.push(token);
        collector = token[4] = [];
        break;
      case '/':
        section = sections.pop();
        section[5] = token[2];
        collector = sections.length > 0 ? sections[sections.length - 1][4] : nestedTokens;
        break;
      default:
        collector.push(token);
    }
  }

  return nestedTokens;
}

/**
 * A simple string scanner that is used by the template parser to find
 * tokens in template strings.
 */
function Scanner(string) {
  this.string = string;
  this.tail = string;
  this.pos = 0;
}

/**
 * Returns `true` if the tail is empty (end of string).
 */
Scanner.prototype.eos = function eos() {
  return this.tail === '';
};

/**
 * Tries to match the given regular expression at the current position.
 * Returns the matched text if it can match, the empty string otherwise.
 */
Scanner.prototype.scan = function scan(re) {
  var match = this.tail.match(re);

  if (!match || match.index !== 0)
    { return ''; }

  var string = match[0];

  this.tail = this.tail.substring(string.length);
  this.pos += string.length;

  return string;
};

/**
 * Skips all text until the given regular expression can be matched. Returns
 * the skipped string, which is the entire tail if no match can be made.
 */
Scanner.prototype.scanUntil = function scanUntil(re) {
  var index = this.tail.search(re), match;

  switch (index) {
    case -1:
      match = this.tail;
      this.tail = '';
      break;
    case 0:
      match = '';
      break;
    default:
      match = this.tail.substring(0, index);
      this.tail = this.tail.substring(index);
  }

  this.pos += match.length;

  return match;
};

/**
 * Represents a rendering context by wrapping a view object and
 * maintaining a reference to the parent context.
 */
function Context(view, parentContext) {
  this.view = view;
  this.cache = { '.': this.view };
  this.parent = parentContext;
}

/**
 * Creates a new context using the given view with this context
 * as the parent.
 */
Context.prototype.push = function push(view) {
  return new Context(view, this);
};

/**
 * Returns the value of the given name in this context, traversing
 * up the context hierarchy if the value is absent in this context's view.
 */
Context.prototype.lookup = function lookup(name) {
  var cache = this.cache;

  var value;
  if (cache.hasOwnProperty(name)) {
    value = cache[name];
  } else {
    var context = this, names, index, lookupHit = false;

    while (context) {
      if (name.indexOf('.') > 0) {
        value = context.view;
        names = name.split('.');
        index = 0;

        /**
         * Using the dot notion path in `name`, we descend through the
         * nested objects.
         *
         * To be certain that the lookup has been successful, we have to
         * check if the last object in the path actually has the property
         * we are looking for. We store the result in `lookupHit`.
         *
         * This is specially necessary for when the value has been set to
         * `undefined` and we want to avoid looking up parent contexts.
         **/
        while (value != null && index < names.length) {
          if (index === names.length - 1)
            { lookupHit = hasProperty(value, names[index]); }

          value = value[names[index++]];
        }
      } else {
        value = context.view[name];
        lookupHit = hasProperty(context.view, name);
      }

      if (lookupHit)
        { break; }

      context = context.parent;
    }

    cache[name] = value;
  }

  if (isFunction(value))
    { value = value.call(this.view); }

  return value;
};

/**
 * A Writer knows how to take a stream of tokens and render them to a
 * string, given a context. It also maintains a cache of templates to
 * avoid the need to parse the same template twice.
 */
function Writer() {
  this.cache = {};
}

/**
 * Clears all cached templates in this writer.
 */
Writer.prototype.clearCache = function clearCache() {
  this.cache = {};
};

/**
 * Parses and caches the given `template` and returns the array of tokens
 * that is generated from the parse.
 */
Writer.prototype.parse = function parse(template, tags) {
  var cache = this.cache;
  var tokens = cache[template];

  if (tokens == null)
    { tokens = cache[template] = parseTemplate(template, tags); }

  return tokens;
};

/**
 * High-level method that is used to render the given `template` with
 * the given `view`.
 *
 * The optional `partials` argument may be an object that contains the
 * names and templates of partials that are used in the template. It may
 * also be a function that is used to load partial templates on the fly
 * that takes a single argument: the name of the partial.
 */
Writer.prototype.render = function render(template, view, partials) {
  var tokens = this.parse(template);
  var context = (view instanceof Context) ? view : new Context(view);
  return this.renderTokens(tokens, context, partials, template);
};

/**
 * Low-level method that renders the given array of `tokens` using
 * the given `context` and `partials`.
 *
 * Note: The `originalTemplate` is only ever used to extract the portion
 * of the original template that was contained in a higher-order section.
 * If the template doesn't use higher-order sections, this argument may
 * be omitted.
 */
Writer.prototype.renderTokens = function renderTokens(tokens, context, partials, originalTemplate) {
  var this$1 = this;

  var buffer = '';

  var token, symbol, value;
  for (var i = 0, numTokens = tokens.length; i < numTokens; ++i) {
    value = undefined;
    token = tokens[i];
    symbol = token[0];

    if (symbol === '#') { value = this$1.renderSection(token, context, partials, originalTemplate); }
    else if (symbol === '^') { value = this$1.renderInverted(token, context, partials, originalTemplate); }
    else if (symbol === '>') { value = this$1.renderPartial(token, context, partials, originalTemplate); }
    else if (symbol === '&') { value = this$1.unescapedValue(token, context); }
    else if (symbol === 'name') { value = this$1.escapedValue(token, context); }
    else if (symbol === 'text') { value = this$1.rawValue(token); }

    if (value !== undefined)
      { buffer += value; }
  }

  return buffer;
};

Writer.prototype.renderSection = function renderSection(token, context, partials, originalTemplate) {
  var this$1 = this;

  var self = this;
  var buffer = '';
  var value = context.lookup(token[1]);

  // This function is used to render an arbitrary template
  // in the current context by higher-order sections.
  function subRender(template) {
    return self.render(template, context, partials);
  }

  if (!value) { return; }

  if (isArray(value)) {
    for (var j = 0, valueLength = value.length; j < valueLength; ++j) {
      buffer += this$1.renderTokens(token[4], context.push(value[j]), partials, originalTemplate);
    }
  } else if (typeof value === 'object' || typeof value === 'string' || typeof value === 'number') {
    buffer += this.renderTokens(token[4], context.push(value), partials, originalTemplate);
  } else if (isFunction(value)) {
    if (typeof originalTemplate !== 'string')
      { throw new Error('Cannot use higher-order sections without the original template'); }

    // Extract the portion of the original template that the section contains.
    value = value.call(context.view, originalTemplate.slice(token[3], token[5]), subRender);

    if (value != null)
      { buffer += value; }
  } else {
    buffer += this.renderTokens(token[4], context, partials, originalTemplate);
  }
  return buffer;
};

Writer.prototype.renderInverted = function renderInverted(token, context, partials, originalTemplate) {
  var value = context.lookup(token[1]);

  // Use JavaScript's definition of falsy. Include empty arrays.
  // See https://github.com/janl/mustache.js/issues/186
  if (!value || (isArray(value) && value.length === 0))
    { return this.renderTokens(token[4], context, partials, originalTemplate); }
};

Writer.prototype.renderPartial = function renderPartial(token, context, partials) {
  if (!partials) { return; }

  var value = isFunction(partials) ? partials(token[1]) : partials[token[1]];
  if (value != null)
    { return this.renderTokens(this.parse(value), context, partials, value); }
};

Writer.prototype.unescapedValue = function unescapedValue(token, context) {
  var value = context.lookup(token[1]);
  if (value != null)
    { return value; }
};

Writer.prototype.escapedValue = function escapedValue(token, context) {
  var value = context.lookup(token[1]);
  if (value != null)
    { return mustache.escape(value); }
};

Writer.prototype.rawValue = function rawValue(token) {
  return token[1];
};

mustache.name = 'mustache.js';
mustache.version = '2.3.0';
mustache.tags = ['{{', '}}'];

// All high-level mustache.* functions use this writer.
var defaultWriter = new Writer();

/**
 * Clears all cached templates in the default writer.
 */
mustache.clearCache = function clearCache() {
  return defaultWriter.clearCache();
};

/**
 * Parses and caches the given template in the default writer and returns the
 * array of tokens it contains. Doing this ahead of time avoids the need to
 * parse templates on the fly as they are rendered.
 */
mustache.parse = function parse(template, tags) {
  return defaultWriter.parse(template, tags);
};

/**
 * Renders the `template` with the given `view` and `partials` using the
 * default writer.
 */
mustache.render = function render(template, view, partials) {
  if (typeof template !== 'string') {
    throw new TypeError('Invalid template! Template should be a "string" ' +
      'but "' + typeStr(template) + '" was given as the first ' +
      'argument for mustache#render(template, view, partials)');
  }

  return defaultWriter.render(template, view, partials);
};

// This is here for backwards compatibility with 0.4.x.,
/*eslint-disable */ // eslint wants camel cased function name
mustache.to_html = function to_html(template, view, partials, send) {
  /*eslint-enable*/

  var result = mustache.render(template, view, partials);

  if (isFunction(send)) {
    send(result);
  } else {
    return result;
  }
};

// Export the escaping function so that the user may override it.
// See https://github.com/janl/mustache.js/issues/244
mustache.escape = escapeHtml;

// Export these mainly for testing, but also for advanced usage.
mustache.Scanner = Scanner;
mustache.Context = Context;
mustache.Writer = Writer;

var builtinChunks = {
  'skinning.vert': 'attribute vec4 a_weights;\nattribute vec4 a_joints;\nuniform sampler2D u_jointsTexture;\nuniform float u_jointsTextureSize;\nmat4 getBoneMatrix(const in float i) {\n  float size = u_jointsTextureSize;\n  float j = i * 4.0;\n  float x = mod(j, size);\n  float y = floor(j / size);\n  float dx = 1.0 / size;\n  float dy = 1.0 / size;\n  y = dy * (y + 0.5);\n  vec4 v1 = texture2D(u_jointsTexture, vec2(dx * (x + 0.5), y));\n  vec4 v2 = texture2D(u_jointsTexture, vec2(dx * (x + 1.5), y));\n  vec4 v3 = texture2D(u_jointsTexture, vec2(dx * (x + 2.5), y));\n  vec4 v4 = texture2D(u_jointsTexture, vec2(dx * (x + 3.5), y));\n  return mat4(v1, v2, v3, v4);\n}\nmat4 skinMatrix() {\n  return\n    getBoneMatrix(a_joints.x) * a_weights.x +\n    getBoneMatrix(a_joints.y) * a_weights.y +\n    getBoneMatrix(a_joints.z) * a_weights.z +\n    getBoneMatrix(a_joints.w) * a_weights.w\n    ;\n}',
  'unpack-normal.frag': 'vec3 unpackNormal(vec4 nmap) {\n  return nmap.xyz * 2.0 - 1.0;\n}',
};

var builtinTemplates = [
  {
    name: 'simple',
    vert: '\nattribute vec3 a_position;\nuniform mat4 model;\nuniform mat4 viewProj;\n{{#useTexture}}\n  attribute vec2 a_uv0;\n  varying vec2 uv0;\n{{/useTexture}}\nvoid main () {\n  vec4 pos = viewProj * model * vec4(a_position, 1);\n  {{#useTexture}}\n    uv0 = a_uv0;\n  {{/useTexture}}\n  gl_Position = pos;\n}',
    frag: '\n{{#useTexture}}\n  uniform sampler2D texture;\n  varying vec2 uv0;\n{{/useTexture}}\n{{#useColor}}\n  uniform vec4 color;\n{{/useColor}}\nvoid main () {\n  vec4 o = vec4(1, 1, 1, 1);\n  {{#useTexture}}\n    o *= texture2D(texture, uv0);\n  {{/useTexture}}\n  {{#useColor}}\n    o *= color;\n  {{/useColor}}\n  if (!gl_FrontFacing) {\n    o.rgb *= 0.5;\n  }\n  gl_FragColor = o;\n}',
    options: [
      { name: 'useTexture', },
      { name: 'useColor', } ],
  } ];

var _shdID = 0;

var ProgramLib = function ProgramLib(device, templates, chunks) {
  var this$1 = this;
  if ( templates === void 0 ) templates = [];
  if ( chunks === void 0 ) chunks = {};

  this._device = device;
  this._precision = "precision highp float;\n";

  // register templates
  this._templates = {};
  for (var i = 0; i < builtinTemplates.length; ++i) {
    var tmpl = builtinTemplates[i];
    this$1.define(tmpl.name, tmpl.vert, tmpl.frag, tmpl.options);
  }
  for (var i$1 = 0; i$1 < templates.length; ++i$1) {
    var tmpl$1 = templates[i$1];
    this$1.define(tmpl$1.name, tmpl$1.vert, tmpl$1.frag, tmpl$1.options);
  }

  // register chunks
  this._chunks = {};
  Object.assign(this._chunks, builtinChunks);
  Object.assign(this._chunks, chunks);

  this._cache = {};
};

/**
 * @param {string} name
 * @param {string} template
 * @param {Array} options
 *
 * @example:
 * programLib.define('foobar', vertTmpl, fragTmpl, [
 *   { name: 'shadow' },
 *   { name: 'lightCount', min: 1, max: 4 }
 * ]);
 */
ProgramLib.prototype.define = function define (name, vert, frag, options) {
  if (this._templates[name]) {
    console.warn(("Failed to define shader " + name + ": already exists."));
    return;
  }

  var id = ++_shdID;

  // calculate option mask offset
  var offset = 0;
  var loop = function ( i ) {
    var op = options[i];
    op._offset = offset;

    var cnt = 1;

    if (op.min !== undefined && op.max !== undefined) {
      cnt = Math.ceil((op.max - op.min) * 0.5);

      op._map = function (value) {
        return (value - this._min) << op._offset;
      }.bind(op);
    } else {
      op._map = function (value) {
        if (value) {
          return 1 << op._offset;
        }
        return 0;
      }.bind(op);
    }

    offset += cnt;

    op._offset = offset;
  };

    for (var i = 0; i < options.length; ++i) loop( i );

  vert = this._precision + vert;
  frag = this._precision + frag;

  // pre-parse the vs and fs template to speed up `mustache.render()` method
  mustache.parse(vert);
  mustache.parse(frag);

  // store it
  this._templates[name] = {
    id: id,
    name: name,
    vert: vert,
    frag: frag,
    options: options
  };
};

/**
 * @param {string} name
 * @param {Object} options
 */
ProgramLib.prototype.getKey = function getKey (name, options) {
  var tmpl = this._templates[name];
  var key = 0;
  for (var i = 0; i < tmpl.options.length; ++i) {
    var tmplOpts = tmpl.options[i];
    var value = options[tmplOpts.name];
    if (value === undefined) {
      continue;
    }

    key |= tmplOpts._map(value);
  }

  return key << 8 | tmpl.id;
};

/**
 * @param {string} name
 * @param {Object} options
 */
ProgramLib.prototype.getProgram = function getProgram (name, options) {
  var key = this.getKey(name, options);
  var program = this._cache[key];
  if (program) {
    return program;
  }

  // get template
  var tmpl = this._templates[name];
  var vert = mustache.render(tmpl.vert, options, this._chunks);
  var frag = mustache.render(tmpl.frag, options, this._chunks);

  program = new gfx.Program(this._device, {
    vert: vert,
    frag: frag
  });
  program.link();
  this._cache[key] = program;

  return program;
};

var _m3_tmp$1 = vmath.mat3.create();
var _m4_tmp$2 = vmath.mat4.create();

var _stageInfos = new memop.RecyclePool(function () {
  return {
    stage: null,
    items: null,
  };
}, 8);

var _float2_pool = new memop.RecyclePool(function () {
  return new Float32Array(2);
}, 8);

var _float3_pool = new memop.RecyclePool(function () {
  return new Float32Array(3);
}, 8);

var _float4_pool = new memop.RecyclePool(function () {
  return new Float32Array(4);
}, 8);

var _float9_pool = new memop.RecyclePool(function () {
  return new Float32Array(9);
}, 8);

var _float16_pool = new memop.RecyclePool(function () {
  return new Float32Array(16);
}, 8);

var _float64_pool = new memop.RecyclePool(function () {
  return new Float32Array(64);
}, 8);

var _int2_pool = new memop.RecyclePool(function () {
  return new Int32Array(2);
}, 8);

var _int3_pool = new memop.RecyclePool(function () {
  return new Int32Array(3);
}, 8);

var _int4_pool = new memop.RecyclePool(function () {
  return new Int32Array(4);
}, 8);

var _int64_pool = new memop.RecyclePool(function () {
  return new Int32Array(64);
}, 8);

var _type2uniformValue = {};
_type2uniformValue[enums.PARAM_INT] = function (value) {
    return value;
  };
_type2uniformValue[enums.PARAM_INT2] = function (value) {
    return vmath.vec2.array(_int2_pool.add(), value);
  };
_type2uniformValue[enums.PARAM_INT3] = function (value) {
    return vmath.vec3.array(_int3_pool.add(), value);
  };
_type2uniformValue[enums.PARAM_INT4] = function (value) {
    return vmath.vec4.array(_int4_pool.add(), value);
  };
_type2uniformValue[enums.PARAM_FLOAT] = function (value) {
    return value;
  };
_type2uniformValue[enums.PARAM_FLOAT2] = function (value) {
    return vmath.vec2.array(_float2_pool.add(), value);
  };
_type2uniformValue[enums.PARAM_FLOAT3] = function (value) {
    return vmath.vec3.array(_float3_pool.add(), value);
  };
_type2uniformValue[enums.PARAM_FLOAT4] = function (value) {
    return vmath.vec4.array(_float4_pool.add(), value);
  };
_type2uniformValue[enums.PARAM_COLOR3] = function (value) {
    return vmath.color3.array(_float3_pool.add(), value);
  };
_type2uniformValue[enums.PARAM_COLOR4] = function (value) {
    return vmath.color4.array(_float4_pool.add(), value);
  };
_type2uniformValue[enums.PARAM_MAT2] = function (value) {
    return vmath.mat2.array(_float4_pool.add(), value);
  };
_type2uniformValue[enums.PARAM_MAT3] = function (value) {
    return vmath.mat3.array(_float9_pool.add(), value);
  };
_type2uniformValue[enums.PARAM_MAT4] = function (value) {
    return vmath.mat4.array(_float16_pool.add(), value);
  };

var _type2uniformArrayValue = {};
_type2uniformArrayValue[enums.PARAM_INT] = {
    func: function func (values) {
      var result = _int64_pool.add();
      for (var i = 0; i < values.length; ++i) {
        result[i] = values[i];
      }
      return result;
    },
    size: 1,
  };
_type2uniformArrayValue[enums.PARAM_INT2] = {
    func: function func (values) {
      var result = _int64_pool.add();
      for (var i = 0; i < values.length; ++i) {
        result[2 * i] = values[i].x;
        result[2 * i + 1] = values[i].y;
      }
      return result;
    },
    size: 2,
  };
_type2uniformArrayValue[enums.PARAM_INT3] = {
    func: undefined,
    size: 3,
  };
_type2uniformArrayValue[enums.PARAM_INT4] = {
    func: function func (values) {
      var result = _int64_pool.add();
      for (var i = 0; i < values.length; ++i) {
        var v = values[i];
        result[4 * i] = v.x;
        result[4 * i + 1] = v.y;
        result[4 * i + 2] = v.z;
        result[4 * i + 3] = v.w;
      }
      return result;
    },
    size: 4,
  };
_type2uniformArrayValue[enums.PARAM_FLOAT] = {
    func: function func (values) {
      var result = _float64_pool.add();
      for (var i = 0; i < values.length; ++i) {
        result[i] = values[i];
      }
      return result;
    },
    size: 1
  };
_type2uniformArrayValue[enums.PARAM_FLOAT2] = {
    func: function func (values) {
      var result = _float64_pool.add();
      for (var i = 0; i < values.length; ++i) {
        result[2 * i] = values[i].x;
        result[2 * i + 1] = values[i].y;
      }
      return result;
    },
    size: 2,
  };
_type2uniformArrayValue[enums.PARAM_FLOAT3] = {
    func: undefined,
    size: 3,
  };
_type2uniformArrayValue[enums.PARAM_FLOAT4] = {
    func: function func (values) {
      var result = _float64_pool.add();
      for (var i = 0; i < values.length; ++i) {
        var v = values[i];
        result[4 * i] = v.x;
        result[4 * i + 1] = v.y;
        result[4 * i + 2] = v.z;
        result[4 * i + 3] = v.w;
      }
      return result;
    },
    size: 4,
  };
_type2uniformArrayValue[enums.PARAM_COLOR3] = {
    func: undefined,
    size: 3,
  };
_type2uniformArrayValue[enums.PARAM_COLOR4] = {
    func: function func (values) {
      var result = _float64_pool.add();
      for (var i = 0; i < values.length; ++i) {
        var v = values[i];
        result[4 * i] = v.r;
        result[4 * i + 1] = v.g;
        result[4 * i + 2] = v.b;
        result[4 * i + 3] = v.a;
      }
      return result;
    },
    size: 4,
  };
_type2uniformArrayValue[enums.PARAM_MAT2] = {
    func: function func (values) {
      var result = _float64_pool.add();
      for (var i = 0; i < values.length; ++i) {
        var v = values[i];
        result[4 * i] = v.m00;
        result[4 * i + 1] = v.m01;
        result[4 * i + 2] = v.m02;
        result[4 * i + 3] = v.m03;
      }
      return result;
    },
    size: 4
  };
_type2uniformArrayValue[enums.PARAM_MAT3] = {
    func: undefined,
    size: 9
  };
_type2uniformArrayValue[enums.PARAM_MAT4] = {
    func: function func (values) {
      var result = _float64_pool.add();
      for (var i = 0; i < values.length; ++i) {
        var v = values[i];
        result[16 * i] = v.m00;
        result[16 * i + 1] = v.m01;
        result[16 * i + 2] = v.m02;
        result[16 * i + 3] = v.m03;
        result[16 * i + 4] = v.m04;
        result[16 * i + 5] = v.m05;
        result[16 * i + 6] = v.m06;
        result[16 * i + 7] = v.m07;
        result[16 * i + 8] = v.m08;
        result[16 * i + 9] = v.m09;
        result[16 * i + 10] = v.m10;
        result[16 * i + 11] = v.m11;
        result[16 * i + 12] = v.m12;
        result[16 * i + 13] = v.m13;
        result[16 * i + 14] = v.m14;
        result[16 * i + 15] = v.m15;
      }
      return result;
    },
    size: 16
  };

var Base = function Base(device, opts) {
  this._device = device;
  this._programLib = new ProgramLib(device, opts.programTemplates, opts.programChunks);
  this._opts = opts;
  this._type2defaultValue = ( obj = {}, obj[enums.PARAM_INT] = 0, obj[enums.PARAM_INT2] = vmath.vec2.new(0, 0), obj[enums.PARAM_INT3] = vmath.vec3.new(0, 0, 0), obj[enums.PARAM_INT4] = vmath.vec4.new(0, 0, 0, 0), obj[enums.PARAM_FLOAT] = 0.0, obj[enums.PARAM_FLOAT2] = vmath.vec2.new(0, 0), obj[enums.PARAM_FLOAT3] = vmath.vec3.new(0, 0, 0), obj[enums.PARAM_FLOAT4] = vmath.vec4.new(0, 0, 0, 0), obj[enums.PARAM_COLOR3] = vmath.color3.new(0, 0, 0), obj[enums.PARAM_COLOR4] = vmath.color4.new(0, 0, 0, 1), obj[enums.PARAM_MAT2] = vmath.mat2.create(), obj[enums.PARAM_MAT3] = vmath.mat3.create(), obj[enums.PARAM_MAT4] = vmath.mat4.create(), obj[enums.PARAM_TEXTURE_2D] = opts.defaultTexture, obj[enums.PARAM_TEXTURE_CUBE] = opts.defaultTextureCube, obj );
  var obj;
  this._stage2fn = {};

  this._viewPools = new memop.RecyclePool(function () {
    return new View();
  }, 8);

  this._drawItemsPools = new memop.RecyclePool(function () {
    return {
      model: null,
      node: null,
      ia: null,
      effect: null,
      options: null,
    };
  }, 100);

  this._stageItemsPools = new memop.RecyclePool(function () {
    return new memop.RecyclePool(function () {
      return {
        model: null,
        node: null,
        ia: null,
        effect: null,
        options: null,
        technique: null,
        sortKey: -1,
      };
    }, 100);
  }, 16);
};

Base.prototype._registerStage = function _registerStage (name, fn) {
  this._stage2fn[name] = fn;
};

Base.prototype._reset = function _reset () {
  this._viewPools.reset();
  this._stageItemsPools.reset();
};

Base.prototype._requestView = function _requestView () {
  return this._viewPools.add();
};

Base.prototype._render = function _render (view, scene) {
    var this$1 = this;

  var device = this._device;

  // setup viewport
  device.setViewport(
    view._rect.x,
    view._rect.y,
    view._rect.w,
    view._rect.h
  );

  // setup clear
  var clearOpts = {};
  if (view._clearFlags & enums.CLEAR_COLOR) {
    clearOpts.color = [
      view._color.r,
      view._color.g,
      view._color.b,
      view._color.a
    ];
  }
  if (view._clearFlags & enums.CLEAR_DEPTH) {
    clearOpts.depth = view._depth;
  }
  if (view._clearFlags & enums.CLEAR_STENCIL) {
    clearOpts.stencil = view._stencil;
  }
  device.clear(clearOpts);

  // get all draw items
  this._drawItemsPools.reset();

  for (var i = 0; i < scene._models.length; ++i) {
    var model = scene._models.data[i];

    // TODO: HACK: filter model by view
    if (view._cullingByID) {
      if (model._viewID !== view._id) {
        continue;
      }
    } else {
      if (model._viewID !== -1) {
        continue;
      }
    }

    for (var m = 0; m < model.drawItemCount; ++m) {
      var drawItem = this$1._drawItemsPools.add();
      model.extractDrawItem(drawItem, m);
    }
  }

  // TODO: update frustum
  // TODO: visbility test
  // frustum.update(view._viewProj);

  // dispatch draw items to different stage
  _stageInfos.reset();

  for (var i$1 = 0; i$1 < view._stages.length; ++i$1) {
    var stage = view._stages[i$1];
    var stageItems = this$1._stageItemsPools.add();
    stageItems.reset();

    for (var j = 0; j < this._drawItemsPools.length; ++j) {
      var drawItem$1 = this$1._drawItemsPools.data[j];
      var tech = drawItem$1.effect.getTechnique(stage);

      if (tech) {
        var stageItem = stageItems.add();
        stageItem.model = drawItem$1.model;
        stageItem.node = drawItem$1.node;
        stageItem.ia = drawItem$1.ia;
        stageItem.effect = drawItem$1.effect;
        stageItem.options = drawItem$1.options;
        stageItem.technique = tech;
        stageItem.sortKey = -1;
      }
    }

    var stageInfo = _stageInfos.add();
    stageInfo.stage = stage;
    stageInfo.items = stageItems;
  }

  // render stages
  for (var i$2 = 0; i$2 < _stageInfos.length; ++i$2) {
    var info = _stageInfos.data[i$2];
    var fn = this$1._stage2fn[info.stage];

    fn(view, info.items);
  }
};

Base.prototype._draw = function _draw (item) {
    var this$1 = this;

  var device = this._device;
  var programLib = this._programLib;
  var node = item.node;
    var ia = item.ia;
    var effect = item.effect;
    var technique = item.technique;
    var options = item.options;

  // reset the pool
  // NOTE: we can use drawCounter optimize this
  // TODO: should be configurable
  _float2_pool.reset();
  _float3_pool.reset();
  _float4_pool.reset();
  _float9_pool.reset();
  _float16_pool.reset();
  _float64_pool.reset();
  _int2_pool.reset();
  _int3_pool.reset();
  _int4_pool.reset();
  _int64_pool.reset();

  // set common uniforms
  // TODO: try commit this depends on effect
  // {
  node.getWorldMatrix(_m4_tmp$2);
  device.setUniform('model', vmath.mat4.array(_float16_pool.add(), _m4_tmp$2));

  vmath.mat3.transpose(_m3_tmp$1, vmath.mat3.invert(_m3_tmp$1, vmath.mat3.fromMat4(_m3_tmp$1, _m4_tmp$2)));
  device.setUniform('normalMatrix', vmath.mat3.array(_float9_pool.add(), _m3_tmp$1));
  // }

  // set technique uniforms
  var slot = 0;
  for (var i = 0; i < technique._parameters.length; ++i) {
    var prop = technique._parameters[i];
    var param = effect.getValue(prop.name);

    if (param === undefined) {
      param = prop.val;
    }

    if (param === undefined) {
      param = this$1._type2defaultValue[prop.type];
    }

    if (param === undefined) {
      console.warn(("Failed to set technique property " + (prop.name) + ", value not found."));
      continue;
    }

    if (
      prop.type === enums.PARAM_TEXTURE_2D ||
      prop.type === enums.PARAM_TEXTURE_CUBE
    ) {
      if (prop.size !== undefined) {
        if (prop.size !== param.length) {
          console.error(("The length of texture array (" + (param.length) + ") is not corrent(expect " + (prop.size) + ")."));
          continue;
        }
        var slots = _int64_pool.add();
        for (var index = 0; index < param.length; ++index) {
          slots[index] = slot + index;
        }
        device.setTextureArray(prop.name, param, slots);
        slot = slot + prop.size;
      } else {
        device.setTexture(prop.name, param, slot);
        ++slot;
      }
    } else {
      var convertedValue = (void 0);
      if (prop.size !== undefined) {
        var convertArray = _type2uniformArrayValue[prop.type];
        if (convertArray.func === undefined) {
          console.error('Uniform array of color3/int3/float3/mat3 can not be supportted!');
          continue;
        }
        if (prop.size * convertArray.size > 64) {
          console.error('Uniform array is too long!');
          continue;
        }
        convertedValue = convertArray.func(param);
      } else {
        var convertFn = _type2uniformValue[prop.type];
        convertedValue = convertFn(param);
      }
      device.setUniform(prop.name, convertedValue);
    }
  }

  // for each pass
  for (var i$1 = 0; i$1 < technique._passes.length; ++i$1) {
    var pass = technique._passes[i$1];
    var count = ia.getPrimitiveCount();

    // set vertex buffer
    device.setVertexBuffer(0, ia._vertexBuffer);

    // set index buffer
    if (ia._indexBuffer) {
      device.setIndexBuffer(ia._indexBuffer);
    }

    // set primitive type
    device.setPrimitiveType(ia._primitiveType);

    // set program
    var program = programLib.getProgram(pass._programName, options);
    device.setProgram(program);

    // cull mode
    device.setCullMode(pass._cullMode);

    // blend
    if (pass._blend) {
      device.enableBlend();
      device.setBlendFuncSep(
        pass._blendSrc,
        pass._blendDst,
        pass._blendSrcAlpha,
        pass._blendDstAlpha
      );
      device.setBlendEqSep(
        pass._blendEq,
        pass._blendAlphaEq
      );
      device.setBlendColor32(pass._blendColor);
    }

    // depth test & write
    if (pass._depthTest) {
      device.enableDepthTest();
      device.setDepthFunc(pass._depthFunc);
    }
    if (pass._depthWrite) {
      device.enableDepthWrite();
    }

    // stencil
    if (pass._stencilTest) {
      device.enableStencilTest();

      // front
      device.setStencilFuncFront(
        pass._stencilFuncFront,
        pass._stencilRefFront,
        pass._stencilMaskFront
      );
      device.setStencilOpFront(
        pass._stencilFailOpFront,
        pass._stencilZFailOpFront,
        pass._stencilZPassOpFront,
        pass._stencilWriteMaskFront
      );

      // back
      device.setStencilFuncBack(
        pass._stencilFuncBack,
        pass._stencilRefBack,
        pass._stencilMaskBack
      );
      device.setStencilOpBack(
        pass._stencilFailOpBack,
        pass._stencilZFailOpBack,
        pass._stencilZPassOpBack,
        pass._stencilWriteMaskBack
      );
    }

    // draw pass
    device.draw(ia._start, count);
  }
};

var renderer = {
  // config
  addStage: config.addStage,

  // utils
  createIA: createIA,

  // classes
  Pass: Pass,
  Technique: Technique,
  Effect: Effect,
  InputAssembler: InputAssembler,
  View: View,

  Light: Light,
  Camera: Camera,
  Model: Model,
  Scene: Scene,

  Base: Base,
  ProgramLib: ProgramLib,
};
Object.assign(renderer, enums);

module.exports = renderer;
//# sourceMappingURL=renderer.js.map
